#include "FileIsb.h"

#include <mpma/base/File.h>
#include <fstream>
#include <iomanip>
#include <sstream>

//If defined, a semi-human-readable version of the isb contents will be written to a file upon load.
//#define DUMP_ISB_FILE_INFO

namespace
{
    std::string ParseString(void *data, int len)
    {
        std::string s;
        for (int i=0; i<len; ++i)
        {
            char c=((char*)data)[i];

            if (c>=32 && c<127)
                s.push_back(c);
        }
        return s;
    }

    std::string MakeHex(void *data, int len)
    {
        std::ostringstream s;
        s<<std::hex<<std::setfill('0');
        for (int i=0; i<len; ++i)
            s<<std::setw(2)<<(unsigned int)((unsigned char*)data)[i];
        return s.str();
    }

    std::string FormatAsDebugData(void *data, int len, const std::string &prefix)
    {
        std::string s="\n";
        s+=prefix;
        s+=" Bytes: ";
        s+=MakeHex(data, len);
        s+="\n";
        s+=prefix;
        s+=" Chars: ";
        s+=ParseString(data, len);
        return s;
    }
}

bool FileIsb::Load(const std::string &filename)
{
    Name=filename;

    //open debug stream
#ifdef DUMP_ISB_FILE_INFO
    MPMA::Filename debugFilenameBase(filename);
    std::string  debugLoadFilename="_debugLoad-";
    debugLoadFilename+=debugFilenameBase.GetName(false, true);
    debugLoadFilename+=".txt";
    std::ofstream debugLoad(debugLoadFilename);
#else
    std::stringstream debugLoad;
#endif

    //open file
    std::ifstream srcFile(filename, std::ios::in|std::ios::binary);
    if (!srcFile.is_open())
    {
        debugLoad<<"Failed to open file "<<filename<<"\n";
        return false;
    }

    debugLoad<<filename<<" opened.\n\n";

    //RIFF header block
    char riffHead[4]={0};
    srcFile.read(riffHead, 4);
    debugLoad<<"Header is: "<<ParseString(riffHead, 4)<<"\n";

    uint32 fileSize=0;
    srcFile.read((char*)&fileSize, 4);
    debugLoad<<"File Size: "<<fileSize<<"\n";

    char headerMagic[8]={0};
    srcFile.read(headerMagic, 8);
    debugLoad<<"Header magic value: "<<ParseString(headerMagic, 8)<<"\n";

    uint32 headerTextBlockSize=0;
    srcFile.read((char*)&headerTextBlockSize, 4);
    debugLoad<<"Header Text Block Size: "<<headerTextBlockSize<<"\n";

    if (headerTextBlockSize>0)
    {
        std::vector<char> headerTextBlock;
        headerTextBlock.resize(headerTextBlockSize);
        srcFile.read(&headerTextBlock[0], headerTextBlockSize);
        debugLoad<<"Header Text Block: "<<ParseString(&headerTextBlock[0], headerTextBlock.size())<<"\n";
    }

    //read main blocks
    while (srcFile)
    {
        char blockNameChars[4]={0};
        srcFile.read(blockNameChars, 4);
        if (!srcFile)
            break;
        std::string blockName=ParseString(blockNameChars, 4);
        debugLoad<<"\nBlock Name: "<<blockName<<"\n";

        uint32 blockSize=0;
        srcFile.read((char*)&blockSize, 4);
        debugLoad<<"Block Size: "<<blockSize<<"\n";

        std::vector<unsigned char> blockData;
        if (blockSize>0)
        {
            blockData.resize(blockSize);
            srcFile.read((char*)&blockData[0], blockSize);
        }

        if (blockData.size()>0 && blockName!="LIST" && blockName!="ctdx")
        {
            debugLoad<<"Data bytes: "<<MakeHex(&blockData[0], blockData.size())<<"\n";
            debugLoad<<"Data chars: "<<ParseString(&blockData[0], blockData.size())<<"\n";
        }

        if (blockName=="LIST")
        {
            uint32 listDataIndex=0;

            //starts with an 8 byte "title" of some sort followed by a size of identifier and identifier
            if (listDataIndex+8>=blockData.size())
                break;
            std::string listBlockTitle=ParseString(&blockData[listDataIndex], 8);
            listDataIndex+=8;
            debugLoad<<"\nList Block Title: "<<listBlockTitle<<"\n";

            if (listDataIndex+4>=blockData.size())
                break;
            uint32 listBlockIdSize=0;
            memcpy(&listBlockIdSize, &blockData[listDataIndex], 4);
            listDataIndex+=4;
            debugLoad<<"List Block ID Size: "<<listBlockIdSize<<"\n";

            std::string listIdString;
            if (listBlockIdSize>0)
            {
                std::vector<char> listId;
                listId.resize(listBlockIdSize);
                memcpy(&listId[0], &blockData[listDataIndex], listBlockIdSize);
                listDataIndex+=listBlockIdSize;

                debugLoad<<"List Block ID bytes: "<<MakeHex(&listId[0], listId.size())<<"\n";
                listIdString=ParseString(&listId[0], listId.size());
                debugLoad<<"List Block ID chars: "<<listIdString<<"\n";
            }

            //the rest of the data inside the LIST are blocks inside of it
            uint32 listSubBlockIndex=0;
            std::string listSubBlockIsgn;
            std::string listSubBlockTitle;
            uint32 listSubBlockTracks=0;
            std::vector<unsigned char> listSubBlockData;
            std::vector<unsigned char> listSubBlockRcnt;

            while (listDataIndex<blockData.size())
            {
                if (listDataIndex+4>blockData.size())
                    break;
                std::string listBlockName=ParseString(&blockData[listDataIndex], 4);
                listDataIndex+=4;
                debugLoad<<"\n    List Block Name: "<<listBlockName<<"\n";

                if (listDataIndex+4>blockData.size())
                    break;
                uint32 listBlockSize=0;
                memcpy(&listBlockSize, &blockData[listDataIndex], 4);
                listDataIndex+=4;
                debugLoad<<"    List Block Size: "<<listBlockSize<<"\n";

                if (listDataIndex+listBlockSize>blockData.size())
                    break;
                std::vector<unsigned char> listBlockData;
                if (listBlockSize>0)
                {
                    listBlockData.resize(listBlockSize);
                    memcpy(&listBlockData[0], &blockData[listDataIndex], listBlockSize);
                    listDataIndex+=listBlockSize;
                }

                if (listBlockData.size()>50000)
                {
                    debugLoad<<"    List Data bytes omitted due to size.\n";
                }
                else if (listBlockData.size()>0)
                {
                    debugLoad<<"    List Data bytes: "<<MakeHex(&listBlockData[0], listBlockData.size())<<"\n";
                    debugLoad<<"    List Data chars: "<<ParseString(&listBlockData[0], listBlockData.size())<<"\n";
                }

                if (listBlockName=="indx" && listBlockData.size()==4)
                    listSubBlockIndex=*(int*)&listBlockData[0];
                else if (listBlockName=="isgn")
                    listSubBlockIsgn=ParseString(&listBlockData[0], listBlockData.size());
                else if (listBlockName=="data")
                    listSubBlockData=std::move(listBlockData);
                else if (listBlockName=="trks")
                    listSubBlockTracks=*(int*)&listBlockData[0];
                else if (listBlockName=="titl")
                    listSubBlockTitle=ParseString(&listBlockData[0], listBlockData.size());
                else if (listBlockName=="rcnt")
                    listSubBlockRcnt=std::move(listBlockData);
            }

            //build something out of the list block if we recognize it
            if (listBlockTitle=="samptitl") //sound sample
            {
                IsbSoundSample samp;
                samp.Id=listIdString;
                samp.Index=listSubBlockIndex;
                samp.Isgn=listSubBlockIsgn;
                samp.Data=std::move(listSubBlockData);

                SoundSamples.emplace_back(std::move(samp));
            }
            else if (listBlockTitle=="sqobstat")
            {
                IcbSqobStat sqob;
                sqob.Title=listSubBlockTitle;
                sqob.Index=listSubBlockIndex;
                sqob.Isgn=listSubBlockIsgn;
                sqob.TrackCount=listSubBlockTracks;

                //first 32 bytes of data is unknown, then each 12 bytes block after that is some sort of per-track data.
                uint32 sqobStatDataOffset=32;
                for (uint32 t=0; t<sqob.TrackCount; ++t)
                {
                    if (sqobStatDataOffset>=listSubBlockData.size())
                        break;

                    //first 2 bytes of track data is a track index that lines up with the isb index
                    uint16 trackIndex=*(uint16*)&listSubBlockData[sqobStatDataOffset];
                    sqobStatDataOffset+=2;

                    //don't know what the next 10 bytes are
                    sqobStatDataOffset+=10;

                    sqob.TrackIndexes.push_back(trackIndex);
                }

                SqobStats.push_back(sqob);
            }
            else if (listBlockTitle=="sdristat")
            {
                IcbSdriStat sdri;
                sdri.Title=listSubBlockTitle;
                sdri.Index=listSubBlockIndex;
                sdri.Isgn=listSubBlockIsgn;
                sdri.TrackCount=listSubBlockTracks;

                //looks like a repeated sequence of 4 32-bit values.
                //first is 'sqob'.  second is maybe a sqob index.  third is an incrementing value of unknown purpose.  fourth is a low number of unknown purpose.
                uint32 sqobStatRcntOffset=0;
                while (sqobStatRcntOffset+4*4<=listSubBlockRcnt.size())
                {
                    uint32 curInd=sqobStatRcntOffset;
                    sqobStatRcntOffset+=16;

                    char typeChars[5]={0};
                    memcpy(typeChars, &listSubBlockRcnt[curInd], 4);
                    curInd+=4;
                    if (std::string(typeChars)!="sqob") //not sure what to do with anything but sqob here
                        continue;

                    uint32 sqobIndex=0;
                    memcpy(&sqobIndex, &listSubBlockRcnt[curInd], 4);
                    curInd+=4;

                    sdri.SqobIndexes.push_back(sqobIndex);
                }

                SdriStats.push_back(sdri);
            }
            else if (listBlockTitle=="entostat")
            {
                IcbEntoStat ento;
                ento.Title=listSubBlockTitle;
                ento.Index=listSubBlockIndex;
                ento.Isgn=listSubBlockIsgn;
                ento.TrackCount=listSubBlockTracks;

                EntoStats.push_back(ento);
            }
        }
        else if (blockName=="ctdx")
        {
            uint32 ctdxDataIndex=0;

            //4 byte unknown before to the repeating stuff
            if (ctdxDataIndex+4>blockData.size())
                break;
            uint32 unknown0;
            memcpy(&unknown0, &blockData[ctdxDataIndex], 4);
            ctdxDataIndex+=4;
            debugLoad<<FormatAsDebugData(&unknown0, 4, "    ctdx Unknown0")<<"\n";

            //a bunch of these repeat
            while (ctdxDataIndex<blockData.size())
            {
                //next is a null-terminated 15-bit-unicode-looking string, parse out the ascii portion
                std::string ctdxName;
                for (; ctdxDataIndex<blockData.size(); ctdxDataIndex+=2)
                {
                    if (blockData[ctdxDataIndex]==0)
                    {
                        ctdxDataIndex+=2;
                        break;
                    }
                    ctdxName.push_back(blockData[ctdxDataIndex]);
                }
                debugLoad<<"\n    ctdx Name: "<<ctdxName;

                //some sort of variable sized data is next, not sure what yet.  but it always ends in either: sqob, ento, or sdri
                std::vector<unsigned char> unknownData1;
                for (; ctdxDataIndex<blockData.size(); ++ctdxDataIndex)
                {
                    unknownData1.push_back(blockData[ctdxDataIndex]);

                    //for now look for one of the known ending marks
                    if (ctdxDataIndex+4<blockData.size())
                    {
                        std::string mark=ParseString(&blockData[ctdxDataIndex], 4);
                        if (mark=="sqob" || mark=="ento" || mark=="sdri")
                            break;
                    }
                }

                debugLoad<<FormatAsDebugData(&unknownData1[0], unknownData1.size(), "    ctdx Unknown1");

                //next 4 is some sort of type indicator string
                if (ctdxDataIndex+4>blockData.size())
                    break;
                std::string ctdxType=ParseString(&blockData[ctdxDataIndex], 4);
                ctdxDataIndex+=4;
                debugLoad<<"\n    ctdx Type: "<<ctdxType;

                //next is either 4 bytes or 8 bytes... not yet sure how to tell so check if there's the start of a unicode string after the 4 for now
                std::vector<unsigned char> unknownData2;
                for (int i=0; i<4 && ctdxDataIndex<blockData.size(); ++i)
                {
                    unknownData2.push_back(blockData[ctdxDataIndex]);
                    ++ctdxDataIndex;
                }

                if (ctdxDataIndex+2<=blockData.size())
                {
                    if (!(blockData[ctdxDataIndex]>='A' && blockData[ctdxDataIndex]<='z' && blockData[ctdxDataIndex+1]==0)) //not a string, so go ahead and pull 4 more
                    {
                        for (int i=0; i<4 && ctdxDataIndex<blockData.size(); ++i)
                        {
                            unknownData2.push_back(blockData[ctdxDataIndex]);
                            ++ctdxDataIndex;
                        }
                    }
                }

                debugLoad<<FormatAsDebugData(&unknownData2[0], unknownData2.size(), "    ctdx Unknown2")<<"\n";
            }

            debugLoad<<"\n";
        }
    }

    debugLoad<<"\nFinished Load.\n";
    return true;
}
