#include "IsbPlayer.h"

// -- IsbSamplePlayer

IsbSamplePlayer::IsbSamplePlayer(const IsbSoundSample &soundData): sound(soundData)
{
    playerSource=MakeSourceForSample();
    player=AUDIO::Player::Create();
}

IsbSamplePlayer::~IsbSamplePlayer()
{
    Stop();
}

void IsbSamplePlayer::Play()
{
    Stop();
    playerSource->Seek(0);
    player->PlayStreaming(playerSource, AUDIO::FIXED);
}

void IsbSamplePlayer::Stop()
{
    player->Stop();
}

std::shared_ptr<AUDIO::Source> IsbSamplePlayer::MakeSourceForSample()
{
    return std::make_shared<AUDIO::VorbisFileSource>(&sound.Data[0], sound.Data.size());
}

// -- IcbIsbSqobPlayer

SourceCombiner::SourceCombiner(std::vector<std::shared_ptr<AUDIO::Source>> sourcesToCombine)
{
    sources=sourcesToCombine;
    atEnd=false;
}

nuint SourceCombiner::FillData(void *data, nuint maxSamples)
{
    //combine samples from all sources
    nuint maxReturned=0;
    tempSamplesCombine.clear();
    tempSamplesCombine.resize(maxSamples*2);
    tempSamplesSingle.resize(maxSamples*2);
    for (auto s=sources.begin(); s!=sources.end(); ++s)
    {
        nuint samplesTotalForSource=0;
        while (samplesTotalForSource<maxSamples)
        {
            nuint samplesToRequest=maxSamples-samplesTotalForSource;

            nuint samplesReturned=(**s).FillData(&tempSamplesSingle[samplesTotalForSource*2], samplesToRequest);
            samplesTotalForSource+=samplesReturned;

            if (samplesReturned==0)
                break;
        }

        for (nuint i=0; i<samplesTotalForSource*2; ++i)
            tempSamplesCombine[i]+=tempSamplesSingle[i];

        if (samplesTotalForSource>maxReturned)
            maxReturned=samplesTotalForSource;
    }

    //clip and return
    for (nuint i=0; i<maxReturned*2; ++i)
    {
        nsint samp=tempSamplesCombine[i];
        if (samp>32767)
            samp=32767-samp+32767;
        else if (samp<-32766)
            samp=-32766-samp-32766;
        ((sint16*)data)[i]=(sint16)samp;
    }

    if (maxReturned==0)
        atEnd=true;

    return maxReturned;
}

nuint SourceCombiner::GetRemainingSamples() const
{
    if (atEnd)
        return 0;

    return AUDIO::FINITE_DATA;
}

nuint SourceCombiner::GetSampleRate() const
{
    if (sources.empty())
        return 44100;

    return sources[0]->GetSampleRate();
}

void SourceCombiner::Seek(nuint sampleNumber)
{
    atEnd=false;
    for (auto s=sources.begin(); s!=sources.end(); ++s)
        (**s).Seek(sampleNumber);
}


IcbIsbSqobPlayer::IcbIsbSqobPlayer(const FileIsb &isb, const FileIsb &icb)
{
    for (auto sqob=icb.SqobStats.begin(); sqob!=icb.SqobStats.end(); ++sqob)
    {
        IcbSqobTheme theme;
        theme.Sqob=&*sqob;

        for (auto sqobi=sqob->TrackIndexes.begin(); sqobi!=sqob->TrackIndexes.end(); ++sqobi)
        {
            for (auto sampi=isb.SoundSamples.begin(); sampi!=isb.SoundSamples.end(); ++sampi)
            {
                if (*sqobi==sampi->Index)
                {
                    theme.Tracks.push_back(&*sampi);
                    break;
                }
            }
        }

        SqobThemes.push_back(theme);
    }

    player=AUDIO::Player::Create();
}

void IcbIsbSqobPlayer::Play(IcbSqobTheme *theme)
{
    Stop();

    std::shared_ptr<SourceCombiner> curCombiner=MakeSourceForTheme(theme);
    player->PlayStreaming(curCombiner, AUDIO::FIXED);
}

std::shared_ptr<SourceCombiner> IcbIsbSqobPlayer::MakeSourceForTheme(IcbSqobTheme *theme)
{
    //create vorbis sources for all the tracks to play
    std::vector<std::shared_ptr<AUDIO::VorbisFileSource>> vorbisSources;
    for (auto t=theme->Tracks.begin(); t!=theme->Tracks.end(); ++t)
        vorbisSources.emplace_back(std::make_shared<AUDIO::VorbisFileSource>(&(**t).Data[0], (**t).Data.size()));

    //add them to a combiner
    std::vector<std::shared_ptr<AUDIO::Source>> sourcesToUse;
    for (auto vs=vorbisSources.begin(); vs!=vorbisSources.end(); ++vs)
        sourcesToUse.push_back(*vs);

    std::shared_ptr<SourceCombiner> curCombiner=std::make_shared<SourceCombiner>(sourcesToUse);
    combiners.push_back(curCombiner);
    return curCombiner;
}

void IcbIsbSqobPlayer::Stop()
{
    player->Stop();
    combiners.clear();
}

// -- IcbIsbSdriPlayer

SourceChainer::SourceChainer(std::vector<std::shared_ptr<AUDIO::Source>> sourcesToChain)
{
    sources=sourcesToChain;
    atEnd=false;
    curSource=0;
}

nuint SourceChainer::FillData(void *data, nuint maxSamples)
{
    if (curSource>=sources.size())
    {
        atEnd=true;
        return 0;
    }

    nuint returned=sources[curSource]->FillData(data, maxSamples);
    if (returned>0)
        return returned;

    ++curSource;
    if (curSource>=sources.size())
    {
        atEnd=true;
        return 0;
    }

    sources[curSource]->Seek(0);

    return sources[curSource]->FillData(data, maxSamples);
}

nuint SourceChainer::GetRemainingSamples() const
{
    if (atEnd)
        return 0;

    return AUDIO::FINITE_DATA;
}

nuint SourceChainer::GetSampleRate() const
{
    if (sources.empty())
        return 44100;

    return sources[0]->GetSampleRate();
}

void SourceChainer::Seek(nuint sampleNumber)
{
    atEnd=false;
    curSource=0;
    for (auto s=sources.begin(); s!=sources.end(); ++s)
        (**s).Seek(sampleNumber);
}

IcbIsbSdriPlayer::IcbIsbSdriPlayer(const FileIsb &isb, const FileIsb &icb): sqobData(isb, icb)
{
    for (auto sdri=icb.SdriStats.begin(); sdri!=icb.SdriStats.end(); ++sdri)
    {
        IcbSdriTheme theme;
        theme.Sdri=&*sdri;

        SdriThemes.push_back(theme);
    }

    player=AUDIO::Player::Create();
}

void IcbIsbSdriPlayer::Play(IcbSdriTheme *theme)
{
    Stop();

    chainer=MakeSourceForTheme(theme);
    player->PlayStreaming(chainer, AUDIO::FIXED);
}

std::shared_ptr<AUDIO::Source> IcbIsbSdriPlayer::MakeSourceForTheme(IcbSdriTheme *theme)
{
    std::vector<std::shared_ptr<AUDIO::Source>> sqobSources;

    for (auto sqobIndex=theme->Sdri->SqobIndexes.begin(); sqobIndex!=theme->Sdri->SqobIndexes.end(); ++sqobIndex)
    {
        for (auto sqob=sqobData.SqobThemes.begin(); sqob!=sqobData.SqobThemes.end(); ++sqob)
        {
            if (*sqobIndex==sqob->Sqob->Index)
            {
                sqobSources.emplace_back(sqobData.MakeSourceForTheme(&*sqob));
                break;
            }
        }
    }

    return std::make_shared<SourceChainer>(sqobSources);
}

void IcbIsbSdriPlayer::Stop()
{
    player->Stop();
}
