#include <mpma/Setup.h>
#include <mpma/base/DebugRouter.h>
#include <mpma/gfxsetup/GFXSetup.h>
#include <mpma/gfxsetup/GL.h>
#include <mpma/base/File.h>
#include <mpma/base/MiscStuff.h>
#include <mpma/audio/SaveToFile.h>
#include <memory>
#include <string>
#include <sstream>
#include <GL/glu.h>

#include "FileIsb.h"
#include "IsbPlayer.h"
#include "TextMenu.h"
#include "OSFilePicker.h"
#include "ProgressStream.h"

#include "resource.h"

void PickFileScene();
void PickModeScene(FileIsb &isb, FileIsb &icb);
void PlayRawSamplesScene(FileIsb &isb, FileIsb &icb);
void PlaySqobThemeScene(FileIsb &isb, FileIsb &icb);
void PlaySdriThemeScene(FileIsb &isb, FileIsb &icb);
void PlayEntoThemeScene(FileIsb &isb, FileIsb &icb);

const std::string VersionString="v0.6a";
const std::string VorbisCommentPrefix="Generated by VGMusic "+VersionString;

namespace
{
    std::string StripFileNameToName(const std::string &orig)
    {
        MPMA::Filename fn(orig);
        return fn.GetName(false, false);
    }
}

void PickFileScene()
{
    enum FileType
    {
        File,
        Directory
    };

    MPMA::Filename curDir=MPMA::FileUtils::GetWorkingDirectory();

    while (true)
    {
        TextMenu menu;
        menu.Title="Pick file ("+(const std::string&)curDir+")";

        std::vector<std::string> allDirs=curDir.GetSubDirectories();
        menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("\\ .. \\", "..", Directory, 0));
        for (auto i=allDirs.begin(); i!=allDirs.end(); ++i)
        {
            menu.Choices.emplace_back(TextMenuEntry("\\"+*i+"\\", *i, Directory, 0));
        }

        std::vector<std::string> allFiles=curDir.GetFiles();
        for (auto i=allFiles.begin(); i!=allFiles.end(); ++i)
        {
            MPMA::Filename f(*i);
            if (MISC::MakeLower(f.GetNameExtension())=="isb")
            {
                menu.Choices.emplace_back(TextMenuEntry(f.GetName(false, false), f.GetName(false, false), File, 0));
            }
        }

        TextMenuEntry *choice=menu.GetChoice();
        if (!choice)
            return;

        if (choice->ParamInt==Directory)
        {
            if (choice->ParamString=="..")
                curDir.StepOut();
            else
                curDir.StepInto(choice->ParamString);

            continue;
        }

        std::string fileNameBase=curDir.GetName()+"/"+choice->ParamString;

        FileIsb isb;
        std::string fileNameIsb=fileNameBase+".isb";
        isb.Load(fileNameIsb);

        FileIsb icb;
        std::string fileNameIcb=fileNameBase+".icb";
        icb.Load(fileNameIcb);

        PickModeScene(isb, icb);
    }
}

void PickModeScene(FileIsb &isb, FileIsb &icb)
{
    enum PlayMode
    {
        Cancel,
        EntoTheme,
        SdriTheme,
        SqobTheme,
        RawSamples
    };

    while (true)
    {
        TextMenu menu;
        menu.Title="Pick data to view ("+StripFileNameToName(isb.Name)+")";

        menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("Back", "", Cancel, 0));
        menu.Choices.emplace_back(TextMenuEntry("Raw Samples - Usually the sound of one instrument within part of a song.  May also be non-dynamic music and sound effects.", "", RawSamples, 0));
        if (!icb.SqobStats.empty())
            menu.Choices.emplace_back(TextMenuEntry("Sample Combinations (sqob) - Raw Samples that are combined to form small segments of music.", "", SqobTheme, 0));
        if (!icb.SdriStats.empty())
            menu.Choices.emplace_back(TextMenuEntry("Music Themes (sdri) - A series of Sample Combinations played in sequence.  This is usually the actual music.", "", SdriTheme, 0));
        if (!icb.EntoStats.empty())
            menu.Choices.emplace_back(TextMenuEntry("Unknown (ento) - Don't know what this data is for.  Not currently used.", "", EntoTheme, 0));

        TextMenuEntry *choice=menu.GetChoice();
        if (!choice)
            return;
        PlayMode mode=(PlayMode)choice->ParamInt;

        if (mode==EntoTheme)
            PlayEntoThemeScene(isb, icb);
        else if (mode==SdriTheme)
            PlaySdriThemeScene(isb, icb);
        else if (mode==SqobTheme)
            PlaySqobThemeScene(isb, icb);
        else if (mode==RawSamples)
            PlayRawSamplesScene(isb, icb);
        else
            return;
    }
}

void PlayRawSamplesScene(FileIsb &isb, FileIsb &icb)
{
    enum ControlChoice
    {
        NotControl,
        Back,
        SaveVorbis,
        SaveWave,
        SaveVorbisAll,
        SaveWaveAll
    };

    bool didAddSaveOptions=false;
    IsbSoundSample *curSample=0;
    TextMenu menu;
    menu.Title="Pick sample to play ("+StripFileNameToName(isb.Name)+")";
    menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("Back", "", Back, 0));
    menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("Save All as Ogg Vorbis", "", SaveVorbisAll, 0));
    menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("Save All as Wave", "", SaveWaveAll, 0));

    for (auto i=isb.SoundSamples.begin(); i!=isb.SoundSamples.end(); ++i)
    {
        std::stringstream name;
        name<<i->Id<<" (index: "<<i->Index<<", name: "<<i->Isgn<<")";

        menu.Choices.emplace_back(TextMenuEntry(name.str(), "", NotControl, &*i));
    }

    std::shared_ptr<IsbSamplePlayer> trackPlayer;

    while (true)
    {
        TextMenuEntry *choice=menu.GetChoice();
        if (!choice)
            return;

        ControlChoice cc=(ControlChoice)choice->ParamInt;
        if (cc==Back)
            return;
        else if (cc==SaveVorbis)
        {
            std::string fname=InvokeOSFilePicker(true, "Ogg Vorbis", "*.ogg;*.ogm", "ogg", MPMA::Vary(MPMA::Filename(isb.Name).GetName(false, false))+"-sample-"+MPMA::Vary(MPMA::Filename(curSample->Id).GetName(false,false))+".ogg");
            if (fname.size()>0)
            {
                std::shared_ptr<AUDIO::Source> audioSource=trackPlayer->MakeSourceForSample();
                std::shared_ptr<ProgressStream> progress=std::make_shared<ProgressStream>("Saving sample...");
                progress->SetNewSource(fname, audioSource);
                if (AUDIO::SaveSourceToVorbisFileVBR(fname, progress, 0.5f, std::list<std::pair<std::string, std::string>>{{"Comment", VorbisCommentPrefix}}))
                    progress->SetNewSource(std::string());
                else
                    InvokeOSMessageBox("Failed to save file: "+fname, true);
            }
        }
        else if (cc==SaveWave)
        {
            std::string fname=InvokeOSFilePicker(true, "Wave Sound", "*.wav", "wav", MPMA::Vary(MPMA::Filename(isb.Name).GetName(false, false))+"-sample-"+MPMA::Vary(MPMA::Filename(curSample->Id).GetName(false,false))+".wav");
            if (fname.size()>0)
            {
                std::shared_ptr<AUDIO::Source> audioSource=trackPlayer->MakeSourceForSample();
                std::shared_ptr<ProgressStream> progress=std::make_shared<ProgressStream>("Saving sample...");
                progress->SetNewSource(fname, audioSource);
                if (AUDIO::SaveSourceToWaveFile(fname, progress))
                    progress->SetNewSource(std::string());
                else
                    InvokeOSMessageBox("Failed to save file: "+fname, true);
            }
        }
        else if (cc==SaveVorbisAll || cc==SaveWaveAll)
        {
            std::string dirname=InvokeOSDirectoryPicker();
            if (dirname.size()>0)
            {
                std::shared_ptr<ProgressStream> progress=std::make_shared<ProgressStream>("Saving samples...");
                for (auto &sample:isb.SoundSamples)
                {
                    if (!GFX::GetWindowState())
                        break;

                    std::string fname=dirname+"\\"+MPMA::Filename(isb.Name).GetName(false, false)+"-sample-"+MPMA::Filename(sample.Id).GetName(false,false)+(cc==SaveVorbisAll?".ogg":".wav");
                    IsbSamplePlayer exportPlayer(sample);
                    std::shared_ptr<AUDIO::Source> audioSource=exportPlayer.MakeSourceForSample();
                    progress->SetNewSource(fname, audioSource);
                    bool success;
                    if (cc==SaveVorbisAll)
                        success=AUDIO::SaveSourceToVorbisFileVBR(fname, progress, 0.5f, std::list<std::pair<std::string, std::string>>{{"Comment", VorbisCommentPrefix}});
                    else
                        success=AUDIO::SaveSourceToWaveFile(fname, progress);
                }
                progress->SetNewSource(std::string());
            }
        }
        else
        {
            IsbSoundSample *sound=(IsbSoundSample*)choice->ParamPtr;
            if (!sound)
                return;

            trackPlayer=std::make_shared<IsbSamplePlayer>(*sound);
            trackPlayer->Play();
            curSample=sound;

            if (!didAddSaveOptions)
            {
                didAddSaveOptions=true;
                menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("Save as Ogg Vorbis", "", SaveVorbis, 0));
                menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("Save as Wave", "", SaveWave, 0));
            }
        }
    }
}

void PlaySqobThemeScene(FileIsb &isb, FileIsb &icb)
{
    IcbIsbSqobPlayer sqobPlayer(isb, icb);

    enum ControlChoice
    {
        NotControl,
        Back,
        SaveVorbis,
        SaveWave,
        SaveVorbisAll,
        SaveWaveAll
    };

    bool didAddSaveOptions=false;
    IcbSqobTheme *curTheme=0;
    TextMenu menu;
    menu.Title="Pick sample combination to play ("+StripFileNameToName(isb.Name)+")";
    menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("Back", "", Back, 0));
    menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("Save All as Ogg Vorbis", "", SaveVorbisAll, 0));
    menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("Save All as Wave", "", SaveWaveAll, 0));

    for (auto i=sqobPlayer.SqobThemes.begin(); i!=sqobPlayer.SqobThemes.end(); ++i)
    {
        std::stringstream name;
        name<<i->Sqob->Title<<" (index: "<<i->Sqob->Index<<", "<<"tracks: "<<i->Tracks.size()<<")";

        menu.Choices.emplace_back(TextMenuEntry(name.str(), "", NotControl, &*i));
    }

    while (true)
    {
        TextMenuEntry *choice=menu.GetChoice();
        if (!choice)
            return;

        ControlChoice cc=(ControlChoice)choice->ParamInt;
        if (cc==Back)
            return;
        else if (cc==SaveVorbis)
        {
            std::string fname=InvokeOSFilePicker(true, "Ogg Vorbis", "*.ogg;*.ogm", "ogg", MPMA::Vary(MPMA::Filename(isb.Name).GetName(false, false))+"-combination-"+MPMA::Vary(curTheme->Sqob->Title)+".ogg");
            if (fname.size()>0)
            {
                std::shared_ptr<AUDIO::Source> audioSource=sqobPlayer.MakeSourceForTheme(curTheme);
                std::shared_ptr<ProgressStream> progress=std::make_shared<ProgressStream>("Saving combination...");
                progress->SetNewSource(fname, audioSource);
                if (AUDIO::SaveSourceToVorbisFileVBR(fname, progress, 0.5f, std::list<std::pair<std::string, std::string>>{{"Comment", VorbisCommentPrefix}}))
                    progress->SetNewSource(std::string());
                else
                    InvokeOSMessageBox("Failed to save file: "+fname, true);
            }
        }
        else if (cc==SaveWave)
        {
            std::string fname=InvokeOSFilePicker(true, "Wave Sound", "*.wav", "wav", MPMA::Vary(MPMA::Filename(isb.Name).GetName(false, false))+"-combination-"+MPMA::Vary(curTheme->Sqob->Title)+".wav");
            if (fname.size()>0)
            {
                std::shared_ptr<AUDIO::Source> audioSource=sqobPlayer.MakeSourceForTheme(curTheme);
                std::shared_ptr<ProgressStream> progress=std::make_shared<ProgressStream>("Saving combination...");
                progress->SetNewSource(fname, audioSource);
                if (AUDIO::SaveSourceToWaveFile(fname, progress))
                    progress->SetNewSource(std::string());
                else
                    InvokeOSMessageBox("Failed to save file: "+fname, true);
            }
        }
        else if (cc==SaveVorbisAll || cc==SaveWaveAll)
        {
            std::string dirname=InvokeOSDirectoryPicker();
            if (dirname.size()>0)
            {
                std::shared_ptr<ProgressStream> progress=std::make_shared<ProgressStream>("Saving combinations...");
                for (auto &combo:sqobPlayer.SqobThemes)
                {
                    if (!GFX::GetWindowState())
                        break;

                    std::string fname=dirname+"\\"+MPMA::Filename(isb.Name).GetName(false, false)+"-combination-"+combo.Sqob->Title+(cc==SaveVorbisAll?".ogg":".wav");
                    std::shared_ptr<AUDIO::Source> audioSource=sqobPlayer.MakeSourceForTheme(&combo);
                    progress->SetNewSource(fname, audioSource);
                    bool success;
                    if (cc==SaveVorbisAll)
                        success=AUDIO::SaveSourceToVorbisFileVBR(fname, progress, 0.5f, std::list<std::pair<std::string, std::string>>{{"Comment", VorbisCommentPrefix}});
                    else
                        success=AUDIO::SaveSourceToWaveFile(fname, progress);
                }
                progress->SetNewSource(std::string());
            }
        }
        else
        {
            IcbSqobTheme *theme=(IcbSqobTheme*)choice->ParamPtr;
            if (!theme)
                return;

            sqobPlayer.Play(theme);
            curTheme=theme;

            if (!didAddSaveOptions)
            {
                didAddSaveOptions=true;
                menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("Save as Ogg Vorbis", "", SaveVorbis, 0));
                menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("Save as Wave", "", SaveWave, 0));
            }
        }
    }
}

void PlaySdriThemeScene(FileIsb &isb, FileIsb &icb)
{
    IcbIsbSdriPlayer sdriPlayer(isb, icb);

    enum ControlChoice
    {
        NotControl,
        Back,
        SaveVorbis,
        SaveWave,
        SaveVorbisAll,
        SaveWaveAll
    };

    bool didAddSaveOptions=false;
    IcbSdriTheme *curTheme=0;
    TextMenu menu;
    menu.Title="Pick theme to play ("+StripFileNameToName(isb.Name)+")";
    menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("Back", "", Back, 0));
    menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("Save All as Ogg Vorbis", "", SaveVorbisAll, 0));
    menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("Save All as Wave", "", SaveWaveAll, 0));

    for (auto i=sdriPlayer.SdriThemes.begin(); i!=sdriPlayer.SdriThemes.end(); ++i)
    {
        std::stringstream name;
        name<<i->Sdri->Title<<" (index: "<<i->Sdri->Index<<", "<<"tracks: "<<i->Sdri->TrackCount<<", sequences: "<<i->Sdri->SqobIndexes.size()<<")";

        menu.Choices.emplace_back(TextMenuEntry(name.str(), "", NotControl, &*i));
    }

    while (true)
    {
        TextMenuEntry *choice=menu.GetChoice();
        if (!choice)
            return;

        ControlChoice cc=(ControlChoice)choice->ParamInt;
        if (cc==Back)
            return;
        else if (cc==SaveVorbis)
        {
            std::string fname=InvokeOSFilePicker(true, "Ogg Vorbis", "*.ogg;*.ogm", "ogg", MPMA::Vary(MPMA::Filename(isb.Name).GetName(false, false))+"-theme-"+MPMA::Vary(curTheme->Sdri->Title)+".ogg");
            if (fname.size()>0)
            {
                std::shared_ptr<AUDIO::Source> audioSource=sdriPlayer.MakeSourceForTheme(curTheme);
                std::shared_ptr<ProgressStream> progress=std::make_shared<ProgressStream>("Saving theme...");
                progress->SetNewSource(fname, audioSource);
                if (AUDIO::SaveSourceToVorbisFileVBR(fname, progress, 0.5f, std::list<std::pair<std::string, std::string>>{{"Comment", VorbisCommentPrefix}}))
                    progress->SetNewSource(std::string());
                else
                    InvokeOSMessageBox("Failed to save file: "+fname, true);
            }
        }
        else if (cc==SaveWave)
        {
            std::string fname=InvokeOSFilePicker(true, "Wave Sound", "*.wav", "wav", MPMA::Vary(MPMA::Filename(isb.Name).GetName(false, false))+"-theme-"+MPMA::Vary(curTheme->Sdri->Title)+".wav");
            if (fname.size()>0)
            {
                std::shared_ptr<AUDIO::Source> audioSource=sdriPlayer.MakeSourceForTheme(curTheme);
                std::shared_ptr<ProgressStream> progress=std::make_shared<ProgressStream>("Saving theme...");
                progress->SetNewSource(fname, audioSource);
                if (AUDIO::SaveSourceToWaveFile(fname, progress))
                    progress->SetNewSource(std::string());
                else
                    InvokeOSMessageBox("Failed to save file: "+fname, true);
            }
        }
        else if (cc==SaveVorbisAll || cc==SaveWaveAll)
        {
            std::string dirname=InvokeOSDirectoryPicker();
            if (dirname.size()>0)
            {
                if (!GFX::GetWindowState())
                    break;

                std::shared_ptr<ProgressStream> progress=std::make_shared<ProgressStream>("Saving themes...");
                for (auto &theme:sdriPlayer.SdriThemes)
                {
                    std::string fname=dirname+"\\"+MPMA::Filename(isb.Name).GetName(false, false)+"-theme-"+theme.Sdri->Title+(cc==SaveVorbisAll?".ogg":".wav");
                    std::shared_ptr<AUDIO::Source> audioSource=sdriPlayer.MakeSourceForTheme(&theme);
                    progress->SetNewSource(fname, audioSource);
                    bool success;
                    if (cc==SaveVorbisAll)
                        success=AUDIO::SaveSourceToVorbisFileVBR(fname, progress, 0.5f, std::list<std::pair<std::string, std::string>>{{"Comment", VorbisCommentPrefix}});
                    else
                        success=AUDIO::SaveSourceToWaveFile(fname, progress);
                }
                progress->SetNewSource(std::string());
            }
        }
        else
        {
            IcbSdriTheme *theme=(IcbSdriTheme*)choice->ParamPtr;
            if (!theme)
                return;

            sdriPlayer.Play(theme);
            curTheme=theme;

            if (!didAddSaveOptions)
            {
                didAddSaveOptions=true;
                menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("Save Current as Ogg Vorbis", "", SaveVorbis, 0));
                menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("Save Current as Wave", "", SaveWave, 0));
            }
        }
    }
}

void PlayEntoThemeScene(FileIsb &isb, FileIsb &icb)
{
    TextMenu menu;
    menu.Title="This doesn't do anything yet ("+StripFileNameToName(isb.Name)+")";
    menu.AlwaysVisibleChoices.emplace_back(TextMenuEntry("Back", "", 0, 0));

    for (auto i=icb.EntoStats.begin(); i!=icb.EntoStats.end(); ++i)
    {
        std::stringstream name;
        name<<i->Title<<" (index: "<<i->Index<<", tracks: "<<i->TrackCount<<")";

        menu.Choices.emplace_back(TextMenuEntry(name.str(), "", 0, &*i));
    }

    while (true)
    {
        TextMenuEntry *choice=menu.GetChoice();
        if (!choice)
            return;

        void *theme=choice->ParamPtr;
        if (!theme)
            return;
    }
}

void RealMain()
{
    //write errors to a file
    MPMA::RouterOutputFile fileErrorReport("_error.txt");
    MPMA::ErrorReport().AddOutputMethod(&fileErrorReport);

    //start mpma
    MPMA::InitAndShutdown mpma;

    //init graphics
    GFX::GraphicsSetup gfxsetup;
    gfxsetup.Name="VGMusic - By Luke Lenhart";
    volatile std::string temp=gfxsetup.Name;
    gfxsetup.Width=1024;
    gfxsetup.Height=768;
    gfxsetup.Resizable=false;
    gfxsetup.Name="VGMusic "+VersionString+" - Clawmark says Rar!";
    if (!GFX::SetupWindow(gfxsetup))
        return;

    GFX::WindowIcon icon;
    icon.ResourceIconId=IDI_ICON1;
    GFX::SetWindowIcon(icon);

    if (!glActiveTextureARB)
    {
        MPMA::ErrorReport()<<"\nBasic OpenGL texture extension not supported: glActiveTextureARB.  Exiting.\n";
        return;
    }

    //Go
    PickFileScene();
}
